package main

import (
	"fmt"
	"regexp"
	"sort"
	"strconv"
	"strings"

	"github.com/go-git/go-git/v5/plumbing"
	"github.com/go-git/go-git/v5/plumbing/object"
	"github.com/go-git/go-git/v5/plumbing/storer"

	"github.com/go-git/go-git/v5"
)

var versionRegex = regexp.MustCompile(`^v?(\d+)\.(\d+)\.(\d+).*$`)

type SemanticVersion struct {
	Major int
	Minor int
	Patch int
	Tag   string
}

func (v SemanticVersion) String() string {
	return fmt.Sprintf("%d.%d.%d", v.Major, v.Minor, v.Patch)
}

func ParseSemanticVersion(tag string) SemanticVersion {
	matches := versionRegex.FindStringSubmatch(tag)
	major, _ := strconv.Atoi(matches[1])
	minor, _ := strconv.Atoi(matches[2])
	patch, _ := strconv.Atoi(matches[3])
	return SemanticVersion{major, minor, patch, tag}
}

func (v SemanticVersion) IsLessThan(w SemanticVersion) bool {
	if v.Major != w.Major {
		return v.Major < w.Major
	}
	if v.Minor != w.Minor {
		return v.Minor < w.Minor
	}
	if v.Patch != w.Patch {
		return v.Patch < w.Patch
	}
	return false
}

type CommitType int

const (
	OtherCommit CommitType = iota
	FeatCommit
	BreakingCommit
	FixCommit
)

func (c CommitType) String() string {
	switch c {
	case FeatCommit:
		return "feat"
	case BreakingCommit:
		return "breaking"
	case FixCommit:
		return "fix"
	default:
		return "other"
	}
}

func GetCommitType() (CommitType, error) {

	latestTag, err := getLatestSemanticTag()
	if err != nil {
		return OtherCommit, err
	}

	if verbosity {
		fmt.Println("latest found tag:", latestTag)
	}

	tagCommit, err := getTagCommit(latestTag.Tag)
	if err != nil {
		return OtherCommit, err
	}

	if verbosity {
		fmt.Println("tag commit:", tagCommit)
	}

	commitType, err := getCommitTypeSinceTag(tagCommit)
	if err != nil {
		return OtherCommit, err
	}

	if verbosity {
		fmt.Println("commit type:", commitType)
	}

	return commitType, nil
}

func getLatestSemanticTag() (SemanticVersion, error) {
	tagList, err := getSemanticTags()
	if err != nil {
		return SemanticVersion{}, err
	}

	if len(tagList) == 0 {
		return SemanticVersion{}, fmt.Errorf("no semantic tags found")
	}

	return tagList[len(tagList)-1], nil
}

func getTagCommit(tag string) (*object.Commit, error) {

	var r = gitRepo

	tags, err := r.Tags()
	if err != nil {
		return nil, fmt.Errorf("failed to get tags: %v", err)
	}

	var tagCommit *object.Commit
	err = tags.ForEach(func(t *plumbing.Reference) error {
		if t.Name().Short() == tag {

			tagObj, err := r.TagObject(t.Hash())
			if err == plumbing.ErrObjectNotFound {
				commit, err := r.CommitObject(t.Hash())
				if err != nil {
					return err
				}
				tagCommit = commit
			} else if err != nil {
				return err
			} else {

				tagCommit, err = tagObj.Commit()
				if err != nil {
					return err
				}
			}
			return ErrStopIteration
		}
		return nil
	})

	if err != nil && err != ErrStopIteration {
		return nil, fmt.Errorf("failed to iterate over tags: %v", err)
	}

	if tagCommit == nil {
		return nil, fmt.Errorf("tag '%s' not found in commit log", tag)
	}

	return tagCommit, nil
}

func getSemanticTags() ([]SemanticVersion, error) {

	r := gitRepo

	tags, err := r.Tags()
	if err != nil {
		return nil, fmt.Errorf("failed to get tags: %v", err)
	}

	var tagList []SemanticVersion
	err = tags.ForEach(func(tag *plumbing.Reference) error {
		tagName := tag.Name().Short()

		if verbosity {
			fmt.Println("found tag: ", tagName)
		}

		if versionRegex.MatchString(tagName) {
			tagList = append(tagList, ParseSemanticVersion(tagName))
		}

		return nil
	})

	if err != nil {
		return nil, fmt.Errorf("failed to iterate over tags: %v", err)
	}

	sort.Slice(tagList, func(i, j int) bool {
		return tagList[i].IsLessThan(tagList[j])
	})

	if verbosity {
		fmt.Println("tag list: ", tagList)
	}

	return tagList, nil
}

func getCommitTypeSinceTag(tagCommit *object.Commit) (CommitType, error) {

	r := gitRepo

	cIter, err := r.Log(&git.LogOptions{})
	if err != nil {
		return OtherCommit, fmt.Errorf("failed to get commit log: %v", err)
	}

	var commitType CommitType
	found := false

	counter := 0

	err = cIter.ForEach(func(commit *object.Commit) error {

		counter++

		if commit.Hash == tagCommit.Hash {
			found = true

			if verbosity {
				fmt.Println("found tag commit after", counter, "commits")
			}

			return storer.ErrStop // stop iteration
		}

		message := strings.TrimSpace(commit.Message)

		if strings.HasPrefix(message, "feat") {

			commitType = FeatCommit
			found = true

			if verbosity {
				fmt.Println("found feat commit after", counter, "commits")
			}

		} else if strings.HasPrefix(message, "fix") {

			// if we already found a feat or breaking commit, we don't care about fix commits
			if commitType < FixCommit && commitType > OtherCommit {
				return nil
			}

			commitType = FixCommit

			if verbosity {
				fmt.Println("found fix commit after", counter, "commits")
			}

		} else if containsBreakingChangeFooter(message) {
			commitType = BreakingCommit
			found = true

			if verbosity {
				fmt.Println("found breaking commit after", counter, "commits")
			}

			return storer.ErrStop // stop iteration
		}

		return nil
	})

	if err != nil {
		return OtherCommit, fmt.Errorf("failed to iterate over commit log: %v", err)
	}

	if !found {
		return OtherCommit, fmt.Errorf("tag commit not found in commit log")
	}

	return commitType, nil
}

func containsBreakingChangeFooter(message string) bool {
	lines := strings.Split(message, "\n")
	for i := len(lines) - 1; i >= 0; i-- {
		if strings.TrimSpace(lines[i]) == "" {
			break
		}
		if strings.Contains(strings.ToLower(lines[i]), "breaking change:") {
			return true
		}
	}
	return false
}
